"""Utilities for working with BCO models.

We don't have a lot of functional code in this schema module so this may belong
somewhere else in another galaxy-data package.
"""

import hashlib
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from galaxy.schema.bco import (
    BioComputeObject,
    BioComputeObjectCore,
    ContributionEnum,
    Contributor,
    ExecutionDomain,
    ExternalDataEndpoint,
    GalaxyExtension,
    GalaxyExtensionDomain,
    SoftwarePrerequisite,
)
from galaxy.schema.bco.execution_domain import (
    ScriptItem,
    Uri as ExecutionUri,
)
from galaxy.util.path import StrOrBytesPath


def _galaxy_script_item(galaxy_url: str) -> ScriptItem:
    """Convert a Galaxy URL into a 'ScriptItem'."""
    return ScriptItem(uri=ExecutionUri(uri=galaxy_url))


def extension_domains(galaxy_url: str, galaxy_version: str) -> List[GalaxyExtensionDomain]:
    galaxy_extension = GalaxyExtension(
        galaxy_url=galaxy_url,
        galaxy_version=galaxy_version,
    )
    return [GalaxyExtensionDomain(galaxy_extension=galaxy_extension)]


def galaxy_execution_domain(
    url: str,
    workflow_url: str,
    software_prerequisites: Optional[List[SoftwarePrerequisite]] = None,
    environment_variables: Optional[Dict[str, str]] = None,
) -> ExecutionDomain:
    data_endpoint = ExternalDataEndpoint(
        name="Access to Galaxy",
        url=url,
    )
    execution_domain = ExecutionDomain(
        script=[_galaxy_script_item(workflow_url)],
        script_driver="Galaxy",
        software_prerequisites=software_prerequisites or [],
        external_data_endpoints=[data_endpoint],
        environment_variables=environment_variables or {},
    )
    return execution_domain


def write_to_file(object_id: str, core_object: BioComputeObjectCore, output: StrOrBytesPath) -> None:
    # core_dict = core_object.dict()
    etag = hashlib.sha256(core_object.model_dump_json(indent=4).encode()).hexdigest()
    object = BioComputeObject(object_id=object_id, etag=etag, **core_object.model_dump())
    with open(output, "w") as f:
        f.write(object.model_dump_json(indent=4, exclude_none=True))


def get_contributors(creator_metadata: Optional[List[Dict[str, Any]]]):
    contributors: List[Contributor] = []
    if creator_metadata:
        for creator in creator_metadata:
            name = None
            email = None
            orcid = None

            if creator["class"] == "Person":
                if "name" in creator:
                    name = creator["name"]
                if "email" in creator:
                    email = creator["email"]
                if "identifier" in creator and "://" in creator["identifier"]:  # Must be an URL
                    orcid = creator["identifier"]

            if name:
                contributors.append(
                    Contributor(
                        contribution=[ContributionEnum.contributedBy],
                        name=name,
                        email=email,
                        orcid=orcid,
                    )
                )
    return contributors



from datetime import datetime
from typing import (
    Optional,
    Union,
)

from pydantic import (
    ConfigDict,
    Field,
    RootModel,
)
from typing_extensions import Literal

from galaxy.schema.fields import (
    DecodedDatabaseIdField,
    EncodedDatabaseIdField,
    ModelClassField,
)
from galaxy.schema.schema import (
    CreateTimeField,
    Model,
    SanitizedString,
    TagCollection,
    UpdateTimeField,
    WithModelClass,
)

VisualizationSortByEnum = Literal["create_time", "title", "update_time", "username"]

VISUALIZATION_MODEL_CLASS = Literal["Visualization"]
VISUALIZATION_REVISION_MODEL_CLASS = Literal["VisualizationRevision"]


class VisualizationIndexQueryPayload(Model):
    deleted: bool = False
    show_own: Optional[bool] = None
    show_published: Optional[bool] = None
    show_shared: Optional[bool] = None
    user_id: Optional[DecodedDatabaseIdField] = None
    sort_by: VisualizationSortByEnum = Field(
        "update_time", title="Sort By", description="Sort pages by this attribute."
    )
    sort_desc: Optional[bool] = Field(default=True, title="Sort descending", description="Sort in descending order.")
    search: Optional[str] = Field(default=None, title="Filter text", description="Freetext to search.")
    limit: Optional[int] = Field(default=100, lt=1000, title="Limit", description="Maximum number of pages to return.")
    offset: Optional[int] = Field(default=0, title="Offset", description="Number of pages to skip.")


class VisualizationSummary(Model):
    id: EncodedDatabaseIdField = Field(
        ...,
        title="ID",
        description="Encoded ID of the Visualization.",
    )
    annotation: Optional[str] = Field(
        default=None,
        title="Annotation",
        description="The annotation of this Visualization.",
    )
    dbkey: Optional[str] = Field(
        default=None,
        title="DbKey",
        description="The database key of the visualization.",
    )
    deleted: bool = Field(
        ...,  # Required
        title="Deleted",
        description="Whether this Visualization has been deleted.",
    )
    importable: bool = Field(
        ...,  # Required
        title="Importable",
        description="Whether this Visualization can be imported.",
    )
    published: bool = Field(
        ...,  # Required
        title="Published",
        description="Whether this Visualization has been published.",
    )
    tags: Optional[TagCollection] = Field(
        ...,
        title="Tags",
        description="A list of tags to add to this item.",
    )
    title: str = Field(
        title="Title",
        description="The name of the visualization.",
    )
    type: str = Field(
        ...,
        title="Type",
        description="The type of the visualization.",
    )
    username: str = Field(
        ...,  # Required
        title="Username",
        description="The name of the user owning this Visualization.",
    )
    create_time: Optional[datetime] = CreateTimeField
    update_time: Optional[datetime] = UpdateTimeField
    model_config = ConfigDict(extra="allow")


class VisualizationSummaryList(RootModel):
    root: list[VisualizationSummary] = Field(
        default=[],
        title="List with detailed information of Visualizations.",
    )


class VisualizationRevisionResponse(Model, WithModelClass):
    model_class: VISUALIZATION_REVISION_MODEL_CLASS = ModelClassField(VISUALIZATION_REVISION_MODEL_CLASS)
    id: EncodedDatabaseIdField = Field(
        ...,
        title="ID",
        description="Encoded ID of the Visualization Revision.",
    )
    visualization_id: EncodedDatabaseIdField = Field(
        ...,
        title="Visualization ID",
        description="Encoded ID of the Visualization.",
    )
    title: str = Field(
        ...,
        title="Title",
        description="The name of the visualization revision.",
    )
    dbkey: Optional[str] = Field(
        None,
        title="DbKey",
        description="The database key of the visualization.",
    )
    config: dict = Field(
        ...,
        title="Config",
        description="The config of the visualization revision.",
    )


class VisualizationPluginResponse(Model):
    name: str = Field(
        ...,
        title="Name",
        description="The name of the plugin.",
    )
    html: str = Field(
        ...,
        title="HTML",
        description="The HTML of the plugin.",
    )
    description: str = Field(
        ...,
        title="Description",
        description="The description of the plugin.",
    )
    logo: Optional[str] = Field(
        None,
        title="Logo",
        description="The logo of the plugin.",
    )
    title: Optional[str] = Field(
        None,
        title="Title",
        description="The title of the plugin.",
    )
    target: str = Field(
        ...,
        title="Target",
        description="The target of the plugin.",
    )
    embeddable: bool = Field(
        ...,
        title="Embeddable",
        description="Whether the plugin is embeddable.",
    )
    entry_point: dict = Field(
        ...,
        title="Entry Point",
        description="The entry point of the plugin.",
    )
    settings: Optional[list[dict]] = Field(
        None,
        title="Settings",
        description="The settings of the plugin.",
    )
    tracks: Optional[list[dict]] = Field(
        None,
        title="Tracks",
        description="The tracks of the plugin.",
    )
    specs: Optional[dict] = Field(
        None,
        title="Specs",
        description="The specs of the plugin.",
    )
    href: str = Field(
        ...,
        title="Href",
        description="The href of the plugin.",
    )


class VisualizationShowResponse(Model, WithModelClass):
    model_class: VISUALIZATION_MODEL_CLASS = ModelClassField(VISUALIZATION_MODEL_CLASS)
    id: EncodedDatabaseIdField = Field(
        ...,
        title="ID",
        description="Encoded ID of the Visualization.",
    )
    title: str = Field(
        ...,
        title="Title",
        description="The name of the visualization.",
    )
    type: str = Field(
        ...,
        title="Type",
        description="The type of the visualization.",
    )
    user_id: EncodedDatabaseIdField = Field(
        ...,
        title="User ID",
        description="The ID of the user owning this Visualization.",
    )
    dbkey: Optional[str] = Field(
        None,
        title="DbKey",
        description="The database key of the visualization.",
    )
    slug: Optional[str] = Field(
        None,
        title="Slug",
        description="The slug of the visualization.",
    )
    latest_revision: VisualizationRevisionResponse = Field(
        ...,
        title="Latest Revision",
        description="The latest revision of this Visualization.",
    )
    revisions: list[EncodedDatabaseIdField] = Field(
        ...,
        title="Revisions",
        description="A list of encoded IDs of the revisions of this Visualization.",
    )
    url: str = Field(
        ...,
        title="URL",
        description="The URL of the visualization.",
    )
    username: str = Field(
        ...,
        title="Username",
        description="The name of the user owning this Visualization.",
    )
    email_hash: str = Field(
        ...,
        title="Email Hash",
        description="The hash of the email of the user owning this Visualization.",
    )
    tags: Optional[TagCollection] = Field(
        None,
        title="Tags",
        description="A list of tags to add to this item.",
    )
    annotation: Optional[str] = Field(
        None,
        title="Annotation",
        description="The annotation of this Visualization.",
    )
    plugin: Optional[VisualizationPluginResponse] = Field(
        None,
        title="Plugin",
        description="The plugin of this Visualization.",
    )


class VisualizationCreateResponse(Model):
    id: EncodedDatabaseIdField = Field(
        ...,
        title="ID",
        description="Encoded ID of the Visualization.",
    )


class VisualizationUpdateResponse(Model):
    id: EncodedDatabaseIdField = Field(
        ...,
        title="ID",
        description="Encoded ID of the Visualization.",
    )
    revision: EncodedDatabaseIdField = Field(
        ...,
        title="Revision",
        description="Encoded ID of the Visualization Revision.",
    )


class VisualizationCreatePayload(Model):
    type: str = Field(
        ...,
        title="Type",
        description="The type of the visualization.",
    )
    title: Optional[SanitizedString] = Field(
        SanitizedString("Untitled Visualization"),
        title="Title",
        description="The name of the visualization.",
        min_length=3,
    )
    dbkey: Optional[SanitizedString] = Field(
        None,
        title="DbKey",
        description="The database key of the visualization.",
    )
    slug: Optional[SanitizedString] = Field(
        None,
        title="Slug",
        description="The slug of the visualization.",
    )
    annotation: Optional[SanitizedString] = Field(
        None,
        title="Annotation",
        description="The annotation of the visualization.",
    )
    config: Optional[dict] = Field(
        {},
        title="Config",
        description="The config of the visualization.",
    )


class VisualizationUpdatePayload(Model):
    title: Optional[SanitizedString] = Field(
        None,
        title="Title",
        description="The name of the visualization.",
    )
    dbkey: Optional[SanitizedString] = Field(
        None,
        title="DbKey",
        description="The database key of the visualization.",
    )
    deleted: Optional[bool] = Field(
        False,
        title="Deleted",
        description="Whether this Visualization has been deleted.",
    )
    config: Optional[Union[dict, bytes]] = Field(
        {},
        title="Config",
        description="The config of the visualization.",
    )



import json
from typing import (
    Annotated,
    Any,
    Optional,
    Union,
)

from pydantic import (
    AfterValidator,
    Field,
    field_validator,
    UUID4,
)

from galaxy.schema.schema import (
    AnnotationField,
    InputDataCollectionStep,
    InputDataStep,
    InputParameterStep,
    Model,
    Organization,
    PauseStep,
    Person,
    StoredWorkflowSummary,
    SubworkflowStep,
    ToolStep,
    WorkflowInput,
)

TargetHistoryIdField = Field(
    None,
    title="History ID",
    # description="The history to import the workflow into.",
    description="The encoded history id into which to import.",
)
INPUTS_BY_DESCRIPTION = (
    "How the 'inputs' field maps its inputs (datasets/collections/step parameters) to workflows steps."
)
STEP_PARAMETERS_NORMALIZED_TITLE = "Legacy Step Parameters Normalized"
STEP_PARAMETERS_NORMALIZED_DESCRIPTION = "Indicates if legacy parameters are already normalized to be indexed by the order_index and are specified as a dictionary per step. Legacy-style parameters could previously be specified as one parameter per step or by tool ID."
STEP_PARAMETERS_TITLE = "Legacy Step Parameters"
STEP_PARAMETERS_DESCRIPTION = "Parameters specified per-step for the workflow invocation, this is legacy and you should generally use inputs and only specify the formal parameters of a workflow instead."
ReplacementParametersField = Field(
    None,
    title="Replacement Parameters",
    description="Class of parameters mostly used for string replacement in PJAs. In best practice workflows, these should be replaced with input parameters",
)
UseCachedJobField = Field(
    False,
    title="Use cached job",
    description="Indicated whether to use a cached job for workflow invocation.",
)
PreferredObjectStoreIdField = Field(
    default=None,
    title="Preferred Object Store ID",
    description="The ID of the object store that should be used to store all datasets (can instead specify object store IDs for intermediate and outputs datasts separately) -  - Galaxy's job configuration may override this in some cases but this workflow preference will override tool and user preferences",
)
PreferredIntermediateObjectStoreIdField = Field(
    None,
    title="Preferred Intermediate Object Store ID",
    description="The ID of the object store that should be used to store the intermediate datasets of this workflow -  - Galaxy's job configuration may override this in some cases but this workflow preference will override tool and user preferences",
)
PreferredOutputsObjectStoreIdField = Field(
    None,
    title="Preferred Outputs Object Store ID",
    description="The ID of the object store that should be used to store the marked output datasets of this workflow - Galaxy's job configuration may override this in some cases but this workflow preference will override tool and user preferences.",
)
ResourceParametersField = Field(
    None,
    title="Resource Parameters",
    description="If a workflow_resource_params_file file is defined and the target workflow is configured to consumer resource parameters, they can be specified with this parameter. See https://github.com/galaxyproject/galaxy/pull/4830 for more information.",
)

VALID_INPUTS_BY_ITEMS = ["step_id", "step_index", "step_uuid", "name"]


def validateInputsBy(inputsBy: Optional[str]) -> Optional[str]:
    if inputsBy is not None:
        if not isinstance(inputsBy, str):
            raise ValueError(f"Invalid type for inputsBy {inputsBy}")
        inputsByArray: list[str] = inputsBy.split("|")
        for inputsByItem in inputsByArray:
            if inputsByItem not in VALID_INPUTS_BY_ITEMS:
                raise ValueError(f"Invalid inputsBy delineation {inputsByItem}")
    return inputsBy


InputsByValidator = AfterValidator(validateInputsBy)


class GetTargetHistoryPayload(Model):
    # TODO - Are the descriptions correct?
    history: Optional[str] = Field(
        None,
        title="History",
        # description="The encoded history id - passed exactly like this 'hist_id=...' -  to import the workflow into. Or the name of the new history to import the workflow into.",
        description="The encoded history id - passed exactly like this 'hist_id=...' -  into which to import. Or the name of the new history into which to import.",
    )
    history_id: Optional[str] = TargetHistoryIdField
    new_history_name: Optional[str] = Field(
        None,
        title="New History Name",
        # description="The name of the new history to import the workflow into.",
        description="The name of the new history into which to import.",
    )


class InvokeWorkflowPayload(GetTargetHistoryPayload):
    # TODO - Are the descriptions correct?
    version: Optional[int] = Field(
        None,
        title="Version",
        description="The version of the workflow to invoke.",
    )
    instance: Optional[bool] = Field(
        False,
        title="Is instance",
        description="True when fetching by Workflow ID, False when fetching by StoredWorkflow ID",
    )
    scheduler: Optional[str] = Field(
        None,
        title="Scheduler",
        description="Scheduler to use for workflow invocation.",
    )
    batch: Optional[bool] = Field(
        False,
        title="Batch",
        description="Indicates if the workflow is invoked as a batch.",
    )
    require_exact_tool_versions: Optional[bool] = Field(
        True,
        title="Require Exact Tool Versions",
        description="If true, exact tool versions are required for workflow invocation.",
        # description="TODO",
    )
    allow_tool_state_corrections: Optional[bool] = Field(
        False,
        title="Allow tool state corrections",
        description="Indicates if tool state corrections are allowed for workflow invocation.",
    )
    landing_uuid: Optional[UUID4] = Field(
        None,
        title="Landing UUID",
        description="The UUID of the workflow landing request associated with this invocation.",
    )
    use_cached_job: Optional[bool] = UseCachedJobField
    parameters_normalized: Optional[bool] = Field(
        False,
        title=STEP_PARAMETERS_NORMALIZED_TITLE,
        description=STEP_PARAMETERS_NORMALIZED_DESCRIPTION,
    )

    @field_validator(
        "parameters",
        "inputs",
        "ds_map",
        "resource_params",
        "replacement_params",
        mode="before",
        check_fields=False,
    )
    @classmethod
    def inputs_string_to_json(cls, v):
        if isinstance(v, str):
            return json.loads(v)
        return v

    parameters: Optional[dict[str, Any]] = Field(
        {},
        title=STEP_PARAMETERS_TITLE,
        description=STEP_PARAMETERS_DESCRIPTION,
    )
    inputs: Optional[dict[str, Any]] = Field(
        None,
        title="Inputs",
        description="Specify values for formal inputs to the workflow",
    )
    ds_map: Optional[dict[str, dict[str, Any]]] = Field(
        {},
        title="Legacy Dataset Map",
        description="An older alternative to specifying inputs using database IDs, do not use this and use inputs instead",
        deprecated=True,
    )
    resource_params: Optional[dict[str, Any]] = ResourceParametersField
    replacement_params: Optional[dict[str, Any]] = ReplacementParametersField
    no_add_to_history: Optional[bool] = Field(
        False,
        title="No Add to History",
        description="Indicates if the workflow invocation should not be added to the history.",
    )
    legacy: Optional[bool] = Field(
        False,
        title="Legacy",
        description="Indicating if to use legacy workflow invocation.",
    )
    inputs_by: Annotated[Optional[str], InputsByValidator] = Field(
        None,
        title="Inputs By",
        # lib/galaxy/workflow/run_request.py - see line 60
        description=INPUTS_BY_DESCRIPTION,
    )
    effective_outputs: Optional[Any] = Field(
        None,
        title="Effective Outputs",
        # lib/galaxy/workflow/run_request.py - see line 455
        description="TODO",
    )
    preferred_object_store_id: Optional[str] = PreferredObjectStoreIdField
    preferred_intermediate_object_store_id: Optional[str] = PreferredIntermediateObjectStoreIdField
    preferred_outputs_object_store_id: Optional[str] = PreferredOutputsObjectStoreIdField


class StoredWorkflowDetailed(StoredWorkflowSummary):
    annotation: Optional[str] = AnnotationField  # Inconsistency? See comment on StoredWorkflowSummary.annotations
    license: Optional[str] = Field(
        None, title="License", description="SPDX Identifier of the license associated with this workflow."
    )
    version: int = Field(
        ..., title="Version", description="The version of the workflow represented by an incremental number."
    )
    inputs: dict[int, WorkflowInput] = Field(
        {}, title="Inputs", description="A dictionary containing information about all the inputs of the workflow."
    )
    creator: Optional[list[Union[Person, Organization]]] = Field(
        None,
        title="Creator",
        description=("Additional information about the creator (or multiple creators) of this workflow."),
    )
    creator_deleted: bool = Field(
        ...,
        title="Creator deleted",
        description="Whether the creator of this Workflow has been deleted.",
    )
    doi: Optional[list[str]] = Field(
        None, title="DOI", description="A list of Digital Object Identifiers associated with this workflow."
    )
    steps: dict[
        int,
        Annotated[
            Union[
                InputDataStep,
                InputDataCollectionStep,
                InputParameterStep,
                PauseStep,
                ToolStep,
                SubworkflowStep,
            ],
            Field(discriminator="type"),
        ],
    ] = Field(
        {},
        title="Steps",
        description="A dictionary with information about all the steps of the workflow.",
    )
    importable: Optional[bool] = Field(
        ...,
        title="Importable",
        description="Indicates if the workflow is importable by the current user.",
    )
    email_hash: Optional[str] = Field(
        ...,
        title="Email Hash",
        description="The hash of the email of the creator of this workflow",
    )
    readme: Optional[str] = Field(
        ...,
        title="Readme",
        description="The detailed markdown readme of the workflow.",
    )
    help: Optional[str] = Field(
        ...,
        title="Help",
        description="The detailed help text for how to use the workflow and debug problems with it.",
    )
    slug: Optional[str] = Field(
        ...,
        title="Slug",
        description="The slug of the workflow.",
    )
    source_metadata: Optional[dict[str, Any]] = Field(
        ...,
        title="Source Metadata",
        description="The source metadata of the workflow.",
    )



from typing import (
    Annotated,
    Any,
    Optional,
)

from pydantic import (
    ConfigDict,
    Field,
)

from galaxy.schema.schema import Model

# Schema manually typed from https://docs.discourse.org/
# Discourse has an OpenApi schema (https://docs.discourse.org/openapi.json)
# but at the time of writing it is mostly untyped, which makes the manually
# created types preferable.


# TODO: remove this class once we have a proper model for all temp models
class HelpTempBaseModel(Model):
    model_config = ConfigDict(extra="allow")


class HelpForumPost(HelpTempBaseModel):
    """Model for a post in the help forum."""

    id: Annotated[int, Field(description="The ID of the post.")]
    name: Annotated[Optional[str], Field(description="The name of the post.")]
    username: Annotated[Optional[str], Field(description="The username of the post author.")]
    avatar_template: Annotated[Optional[str], Field(description="The avatar template of the user.")]
    created_at: Annotated[Optional[str], Field(description="The creation date of the post.")]
    like_count: Annotated[Optional[int], Field(description="The number of likes of the post.")]
    blurb: Annotated[Optional[str], Field(description="The blurb of the post.")]
    post_number: Annotated[Optional[int], Field(description="The post number of the post.")]
    topic_id: Annotated[Optional[int], Field(description="The ID of the topic of the post.")]


class HelpForumTopic(Model):
    """Model for a topic in the help forum compatible with Discourse API."""

    id: Annotated[int, Field(description="The ID of the topic.")]
    title: Annotated[str, Field(description="The title of the topic.")]
    fancy_title: Annotated[str, Field(description="The fancy title of the topic.")]
    slug: Annotated[str, Field(description="The slug of the topic.")]
    posts_count: Annotated[int, Field(description="The number of posts in the topic.")]
    reply_count: Annotated[int, Field(description="The number of replies in the topic.")]
    highest_post_number: Annotated[int, Field(description="The highest post number in the topic.")]
    created_at: Annotated[str, Field(description="The creation date of the topic.")]
    last_posted_at: Annotated[str, Field(description="The date of the last post in the topic.")]
    bumped: Annotated[bool, Field(description="Whether the topic was bumped.")]
    bumped_at: Annotated[str, Field(description="The date of the last bump of the topic.")]
    archetype: Annotated[Any, Field(description="The archetype of the topic.")]
    unseen: Annotated[bool, Field(description="Whether the topic is unseen.")]
    pinned: Annotated[bool, Field(description="Whether the topic is pinned.")]
    unpinned: Annotated[Optional[bool], Field(default=None, description="Whether the topic is unpinned.")]
    visible: Annotated[bool, Field(description="Whether the topic is visible.")]
    closed: Annotated[bool, Field(description="Whether the topic is closed.")]
    archived: Annotated[bool, Field(description="Whether the topic is archived.")]
    bookmarked: Annotated[Optional[bool], Field(default=None, description="Whether the topic is bookmarked.")]
    liked: Annotated[Optional[bool], Field(default=None, description="Whether the topic is liked.")]
    tags: Annotated[list[str], Field(description="The tags of the topic.")]
    tags_descriptions: Annotated[
        Optional[Any], Field(default=None, description="The descriptions of the tags of the topic.")
    ]
    category_id: Annotated[int, Field(description="The ID of the category of the topic.")]
    has_accepted_answer: Annotated[bool, Field(description="Whether the topic has an accepted answer.")]


class HelpForumUser(HelpTempBaseModel):
    """Model for a user in the help forum."""

    pass


class HelpForumCategory(HelpTempBaseModel):
    """Model for a category in the help forum."""

    pass


class HelpForumTag(HelpTempBaseModel):
    """Model for a tag in the help forum."""

    pass


class HelpForumGroup(HelpTempBaseModel):
    """Model for a group in the help forum."""

    pass


class HelpForumGroupedSearchResult(HelpTempBaseModel):
    """Model for a grouped search result."""

    pass


class HelpForumSearchResponse(Model):
    """Response model for the help search API endpoint.

    This model is based on the Discourse API response for the search endpoint.
    """

    posts: Annotated[list[HelpForumPost], Field(default=None, description="The list of posts returned by the search.")]
    topics: Annotated[
        list[HelpForumTopic], Field(default=None, description="The list of topics returned by the search.")
    ]
    users: Annotated[
        Optional[list[HelpForumUser]], Field(default=None, description="The list of users returned by the search.")
    ]
    categories: Annotated[
        Optional[list[HelpForumCategory]],
        Field(default=None, description="The list of categories returned by the search."),
    ]
    tags: Annotated[
        Optional[list[HelpForumTag]], Field(default=None, description="The list of tags returned by the search.")
    ]
    groups: Annotated[
        Optional[list[HelpForumGroup]], Field(default=None, description="The list of groups returned by the search.")
    ]
    grouped_search_result: Annotated[
        Optional[HelpForumGroupedSearchResult], Field(default=None, description="The grouped search result.")
    ]



import sys
from typing import (
    Any,
    Generic,
    TypeVar,
)

from pydantic import BaseModel
from pydantic.json_schema import GenerateJsonSchema
from typing_extensions import override

from galaxy.schema.fields import (
    DecodedDatabaseIdField,
    EncodedDatabaseIdField,
)

DatabaseIdT = TypeVar("DatabaseIdT")

ref_to_name = {}


class GenericModel(BaseModel):
    @classmethod
    def model_parametrized_name(cls, params: tuple[type[Any], ...]) -> str:
        suffix = cls.__determine_suffix__(params)
        class_name = cls.__name__.split("Generic", 1)[-1]
        return f"{class_name}{suffix}"

    @classmethod
    def __get_pydantic_core_schema__(cls, *args, **kwargs):
        result = super().__get_pydantic_core_schema__(*args, **kwargs)
        ref_to_name[result["ref"]] = cls.__name__
        return result

    @classmethod
    def __determine_suffix__(cls, params: tuple[type[Any], ...]) -> str:
        suffix = "Incoming"
        if params[0] is EncodedDatabaseIdField:
            suffix = "Response"
        elif params[0] is DecodedDatabaseIdField:
            suffix = "Request"
        return suffix


class CustomJsonSchema(GenerateJsonSchema):
    def get_defs_ref(self, core_mode_ref):
        full_def = super().get_defs_ref(core_mode_ref)
        choices = self._prioritized_defsref_choices[full_def]
        ref, mode = core_mode_ref
        if ref in ref_to_name:
            for i, choice in enumerate(choices):
                choices[i] = choice.replace(choices[0], ref_to_name[ref])  # type: ignore[call-overload]
        return full_def


class PatchGenericPickle:
    """A mixin that allows generic pydantic models to be serialized and deserialized with pickle.

    Notes
    ----
    In general, pickle shouldn't be encouraged as a means of serialization since there are better,
    safer options. In some cases e.g. Streamlit's `@st.cache_data there's no getting around
    needing to use pickle.

    As of Pydantic 2.7, generics don't properly work with pickle. The core issue is the following
    1. For each specialized generic, pydantic creates a new subclass at runtime. This class
       has a `__qualname__` that contains the type var argument e.g. `"MyGeneric[str]"` for a
       `class MyGeneric(BaseModel, Generic[T])`.
    2. Pickle attempts to find a symbol with the value of `__qualname__` in the module where the
       class was defined, which fails since Pydantic defines that class dynamically at runtime.
       Pydantic does attempt to register these dynamic classes but currently only for classes
       defined at the top-level of the interpreter.

    See Also
    --------
    - https://github.com/pydantic/pydantic/issues/9390
    """

    @classmethod
    @override
    def __init_subclass__(cls, **kwargs):
        # Note: we're still in __init_subclass__, not yet in __pydantic_init_subclass__
        #  not all model_fields are available at this point.
        super().__init_subclass__(**kwargs)

        if not issubclass(cls, BaseModel):
            raise TypeError("PatchGenericPickle can only be used with subclasses of pydantic.BaseModel")
        if not issubclass(cls, Generic):  # type: ignore[unreachable]  # https://github.com/python/mypy/issues/19377
            raise TypeError("PatchGenericPickle can only be used with Generic models")

        qualname = cls.__qualname__  # type: ignore[unreachable]  # https://github.com/python/mypy/issues/19377
        declaring_module = sys.modules[cls.__module__]
        if qualname not in declaring_module.__dict__:
            # This should work in all cases, but we might need to make this check and update more
            # involved e.g. see pydantic._internal._generics.create_generic_submodel
            declaring_module.__dict__[qualname] = cls



# generated by datamodel-codegen:
#   filename:  https://opensource.ieee.org/2791-object/ieee-2791-schema/-/raw/master/io_domain.json
#   timestamp: 2022-09-13T23:51:51+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import (
    List,
    Optional,
)

from pydantic import (
    AnyUrl,
    BaseModel,
    ConfigDict,
    EmailStr,
    Field,
    RootModel,
)


class Uri(BaseModel):
    model_config = ConfigDict(extra="forbid")

    filename: Optional[str] = None
    uri: AnyUrl
    access_time: Optional[datetime] = Field(
        None, description="Time stamp of when the request for this data was submitted"
    )
    sha1_checksum: Optional[str] = Field(
        None, description="output of hash function that produces a message digest", pattern="[A-Za-z0-9]+"
    )


class ObjectId(RootModel):
    root: str = Field(
        ...,
        description="A unique identifier that should be applied to each IEEE-2791 Object instance, generated and assigned by a IEEE-2791 database engine. IDs should never be reused",
    )


class ContributionEnum(Enum):
    authoredBy = "authoredBy"
    contributedBy = "contributedBy"
    createdAt = "createdAt"
    createdBy = "createdBy"
    createdWith = "createdWith"
    curatedBy = "curatedBy"
    derivedFrom = "derivedFrom"
    importedBy = "importedBy"
    importedFrom = "importedFrom"
    providedBy = "providedBy"
    retrievedBy = "retrievedBy"
    retrievedFrom = "retrievedFrom"
    sourceAccessedBy = "sourceAccessedBy"


class Contributor(BaseModel):
    model_config = ConfigDict(extra="forbid")

    name: str = Field(..., description="Name of contributor", examples=["Charles Darwin"])
    affiliation: Optional[str] = Field(
        None, description="Organization the particular contributor is affiliated with", examples=["HMS Beagle"]
    )
    email: Optional[EmailStr] = Field(
        None,
        description="electronic means for identification and communication purposes",
        examples=["name@example.edu"],
    )
    contribution: List[ContributionEnum] = Field(
        ..., description="type of contribution determined according to PAV ontology"
    )
    orcid: Optional[AnyUrl] = Field(
        None,
        description="Field to record author information. ORCID identifiers allow for the author to curate their information after submission. ORCID identifiers must be valid and must have the prefix ‘https://orcid.org/’",
        examples=["http://orcid.org/0000-0002-1825-0097"],
    )


class InputSubdomainItem(BaseModel):
    model_config = ConfigDict(extra="forbid")

    uri: Uri


class OutputSubdomainItem(BaseModel):
    mediatype: str = Field(
        ...,
        description="https://www.iana.org/assignments/media-types/",
        examples=["text/csv"],
        pattern="^(.*)$",
        title="mediatype",
    )
    uri: Uri


class InputAndOutputDomain(BaseModel):
    input_subdomain: List[InputSubdomainItem] = Field(
        ...,
        description="A record of the references and input files for the entire pipeline. Each type of input file is listed under a key for that type.",
        title="input_domain",
    )
    output_subdomain: List[OutputSubdomainItem] = Field(
        ..., description="A record of the outputs for the entire pipeline.", title="output_subdomain"
    )



from datetime import datetime
from typing import (
    Annotated,
    Union,
)

from pydantic import ValidationInfo
from pydantic.functional_validators import AfterValidator
from typing_extensions import (
    Literal,
)

# Relative URLs cannot be validated with AnyUrl, they need a scheme.
# Making them an alias of `str` for now
RelativeUrl = str

# TODO: we may want to add a custom validator for this and for RelativeUrl
AbsoluteOrRelativeUrl = RelativeUrl

LatestLiteral = Literal["latest"]


def strip_tzinfo(v: datetime, info: ValidationInfo) -> datetime:
    if v.tzinfo:
        if offset := v.utcoffset():
            return v.replace(tzinfo=None) - offset
        return v.replace(tzinfo=None)
    return v


OffsetNaiveDatetime = Annotated[datetime, AfterValidator(strip_tzinfo)]

CoercedStringType = Annotated[
    Union[str, int, float, bool], AfterValidator(lambda val: val if isinstance(val, str) else str(val))
]



from typing import List

from pydantic import BaseModel
from typing_extensions import Literal

from .description_domain import (
    DescriptionDomain,
    PipelineStep,
    Uri as DescriptionDomainUri,
    XrefItem,
)
from .error_domain import ErrorDomain
from .execution_domain import (
    ExecutionDomain,
    ExternalDataEndpoint,
    ScriptItem,
    SoftwarePrerequisite,
    Uri as ExecutionDomainUri,
)
from .io_domain import (
    InputAndOutputDomain,
    InputSubdomainItem,
    OutputSubdomainItem,
    Uri as InputAndOutputDomainUri,
)
from .parametric_domain import (
    ParametricDomain,
    ParametricDomainItem,
)
from .provenance_domain import (
    ContributionEnum,
    Contributor,
    ProvenanceDomain,
)
from .usability_domain import UsabilityDomain

SPEC_VERSION = "https://w3id.org/ieee/ieee-2791-schema/2791object.json"


class GalaxyExtension(BaseModel):
    galaxy_url: str
    galaxy_version: str


class GalaxyExtensionDomain(BaseModel):
    extension_schema: Literal["galaxy_extension.json"] = "galaxy_extension.json"
    galaxy_extension: GalaxyExtension


class BioComputeObjectCore(BaseModel):
    """BioComputeObject pre-hash, used to generate etag."""

    description_domain: DescriptionDomain
    error_domain: ErrorDomain
    execution_domain: ExecutionDomain
    extension_domain: List[GalaxyExtensionDomain]
    io_domain: InputAndOutputDomain
    parametric_domain: ParametricDomain
    provenance_domain: ProvenanceDomain
    usability_domain: UsabilityDomain


class BioComputeObject(BioComputeObjectCore):
    object_id: str
    spec_version: str = SPEC_VERSION
    etag: str


__all__ = (
    "BioComputeObjectCore",
    "BioComputeObject",
    "Contributor",
    "ContributionEnum",
    "DescriptionDomain",
    "DescriptionDomainUri",
    "ErrorDomain",
    "ExecutionDomain",
    "ExecutionDomainUri",
    "ExternalDataEndpoint",
    "GalaxyExtension",
    "GalaxyExtensionDomain",
    "InputAndOutputDomain",
    "InputSubdomainItem",
    "OutputSubdomainItem",
    "InputAndOutputDomainUri",
    "ParametricDomain",
    "ParametricDomainItem",
    "PipelineStep",
    "ProvenanceDomain",
    "ScriptItem",
    "SoftwarePrerequisite",
    "SPEC_VERSION",
    "UsabilityDomain",
    "XrefItem",
)



from enum import Enum
from typing import (
    Annotated,
    Any,
    Literal,
    Optional,
    Union,
)

from pydantic import (
    Field,
    RootModel,
)

from galaxy.schema.schema import Model
from galaxy.util.hash_util import HashFunctionNames


class RemoteFilesTarget(str, Enum):
    ftpdir = "ftpdir"
    userdir = "userdir"
    importdir = "importdir"


class RemoteFilesFormat(str, Enum):
    flat = "flat"
    jstree = "jstree"
    uri = "uri"


class RemoteFilesDisableMode(str, Enum):
    folders = "folders"
    files = "files"


class FilesSourceSupports(Model):
    pagination: Annotated[bool, Field(description="Whether this file source supports server-side pagination.")] = False
    search: Annotated[bool, Field(description="Whether this file source supports server-side search.")] = False
    sorting: Annotated[bool, Field(description="Whether this file source supports server-side sorting.")] = False


class FilesSourcePlugin(Model):
    id: str = Field(
        ...,
        title="ID",
        description="The `FilesSource` plugin identifier",
        examples=["_import"],
    )
    type: str = Field(
        ...,
        title="Type",
        description="The type of the plugin.",
        examples=["gximport"],
    )
    label: str = Field(
        ...,
        title="Label",
        description="The display label for this plugin.",
        examples=["Library Import Directory"],
    )
    doc: Optional[str] = Field(
        None,
        title="Documentation",
        description="Documentation or extended description for this plugin.",
        examples=["Galaxy's library import directory"],
    )
    browsable: bool = Field(
        ...,
        title="Browsable",
        description="Whether this file source plugin can list items.",
    )
    writable: bool = Field(
        ...,
        title="Writeable",
        description="Whether this files source plugin allows write access.",
        examples=[False],
    )
    requires_roles: Optional[str] = Field(
        None,
        title="Requires roles",
        description="Only users with the roles specified here can access this files source.",
    )
    requires_groups: Optional[str] = Field(
        None,
        title="Requires groups",
        description="Only users belonging to the groups specified here can access this files source.",
    )
    url: Optional[str] = Field(
        None,
        title="URL",
        description="Optional URL that might be provided by some plugins to link to the remote source.",
    )
    supports: Annotated[
        FilesSourceSupports,
        Field(default=..., description="Features supported by this file source."),
    ] = FilesSourceSupports()


class BrowsableFilesSourcePlugin(FilesSourcePlugin):
    browsable: Literal[True]
    uri_root: str = Field(
        ...,
        title="URI root",
        description="The URI root used by this type of plugin.",
        examples=["gximport://"],
    )


class FilesSourcePluginList(RootModel):
    root: list[Union[BrowsableFilesSourcePlugin, FilesSourcePlugin]] = Field(
        default=[],
        title="List of files source plugins",
        examples=[
            {
                "id": "_import",
                "type": "gximport",
                "uri_root": "gximport://",
                "label": "Library Import Directory",
                "doc": "Galaxy's library import directory",
                "writable": False,
                "browsable": True,
            }
        ],
    )


class RemoteEntry(Model):
    name: str = Field(..., title="Name", description="The name of the entry.")
    uri: str = Field(..., title="URI", description="The URI of the entry.")
    path: str = Field(..., title="Path", description="The path of the entry.")


class RemoteDirectory(RemoteEntry):
    class_: Literal["Directory"] = Field(..., alias="class")


class RemoteFileHash(Model):
    hash_function: HashFunctionNames
    hash_value: str


class RemoteFile(RemoteEntry):
    class_: Literal["File"] = Field(..., alias="class")
    size: int = Field(..., title="Size", description="The size of the file in bytes.")
    ctime: str = Field(..., title="Creation time", description="The creation time of the file.")
    hashes: Optional[list[RemoteFileHash]] = Field(
        None, title="Hashes", description="List of precomputed hashes for the file, if available."
    )


class ListJstreeResponse(RootModel):
    root: list[Any] = Field(
        default=[],
        title="List of files",
        description="List of files in Jstree format.",
        # TODO: also deprecate on python side, https://github.com/pydantic/pydantic/issues/2255
        json_schema_extra={"deprecated": True},
    )


AnyRemoteEntry = Annotated[
    Union[RemoteFile, RemoteDirectory],
    Field(discriminator="class_"),
]


class ListUriResponse(RootModel):
    root: list[AnyRemoteEntry] = Field(
        default=[],
        title="List of remote entries",
        description="List of directories and files.",
    )


AnyRemoteFilesListResponse = Union[ListUriResponse, ListJstreeResponse]


class CreateEntryPayload(Model):
    target: str = Field(
        ...,
        title="Target",
        description="The target file source to create the entry in.",
    )
    name: str = Field(
        ...,
        title="Name",
        description="The name of the entry to create.",
        examples=["my_new_entry"],
    )


class CreatedEntryResponse(Model):
    name: str = Field(
        ...,
        title="Name",
        description="The name of the created entry.",
        examples=["my_new_entry"],
    )
    uri: str = Field(
        ...,
        title="URI",
        description="The URI of the created entry.",
        examples=["gxfiles://my_new_entry"],
    )
    external_link: Optional[str] = Field(
        default=None,
        title="External link",
        description="An optional external link to the created entry if available.",
    )



from typing import Optional

from pydantic import (
    Field,
    RootModel,
)
from typing_extensions import Literal

from galaxy.schema.fields import (
    DecodedDatabaseIdField,
    EncodedDatabaseIdField,
    ModelClassField,
)
from galaxy.schema.schema import (
    Model,
    WithModelClass,
)
from galaxy.util.config_templates import partial_model

GROUP_MODEL_CLASS = Literal["Group"]


class GroupResponse(Model, WithModelClass):
    """Response schema for a group."""

    model_class: GROUP_MODEL_CLASS = ModelClassField(GROUP_MODEL_CLASS)
    id: EncodedDatabaseIdField = Field(
        ...,
        title="group ID",
    )
    name: str = Field(
        ...,
        title="name of the group",
    )
    url: str = Field(
        ...,
        title="URL for the group",
    )
    roles_url: Optional[str] = Field(
        None,
        title="URL for the roles of the group",
    )
    users_url: Optional[str] = Field(
        None,
        title="URL for the users of the group",
    )


class GroupListResponse(RootModel):
    """Response schema for listing groups."""

    root: list[GroupResponse]


class GroupCreatePayload(Model):
    """Payload schema for creating a group."""

    name: str = Field(
        ...,
        title="name of the group",
    )
    user_ids: list[DecodedDatabaseIdField] = Field(
        [],
        title="user IDs",
    )
    role_ids: list[DecodedDatabaseIdField] = Field(
        [],
        title="role IDs",
    )


@partial_model()
class GroupUpdatePayload(Model):
    """Payload schema for updating a group."""

    name: str = Field(
        ...,
        title="name of the group",
    )
    user_ids: Optional[list[DecodedDatabaseIdField]] = Field(
        None,
        title="user IDs",
    )
    role_ids: Optional[list[DecodedDatabaseIdField]] = Field(
        None,
        title="role IDs",
    )
